(
 ******************************************
 * Mark Steere's - Quadrature -           *
 *                                        *
 * Powered by the Axiom meta-game engine. *
 ******************************************
)

$gameLog OFF
$passTurnForced ON
$shareMovesList OFF

{board
	11 11 2DUP {grid} * CONSTANT #BoardSize

	{position} RedMarker
	{position} YellowMarker
board}

{directions
	-1  0 {direction}	North
	 1  0 {direction}	South
	 0  1 {direction}	East
	 0 -1 {direction}	West
	-1  1 {direction}	Northeast
	 1  1 {direction}	Southeast
	-1 -1 {direction}	Northwest
	 1 -1 {direction}	Southwest
directions}


{players
	{player}	Red
	{player}	Yellow
	{neutral}	Referee
players}

{symmetries
	Yellow {symmetry} North South
	Yellow {symmetry} East West
	Yellow {symmetry} Northeast Southwest
	Yellow {symmetry} Northwest Southeast
symmetries}

{turn-order
	{turn}		Red
	{turn}		Yellow
turn-order}


( Forward definition of pieces )
{pieces
	{piece}		Man
	{piece}		RedMarkerI
	{piece}		RedMarkerII
	{piece}		YellowMarkerI
	{piece}		YellowMarkerII
pieces}


\ This is not the best Forth style, but it's fast

VARIABLE pos2Enemy
VARIABLE pos3Enemy

\ Determine if we're are squaring the opponent.
: CheckSquares ( to -- )
	DUP 11 / 11 * DUP 11 + SWAP
	DO
		\ I - pos2
		DUP I <> I not-empty-at? AND
		IF
			I enemy-at? pos2Enemy !
			I 11 MOD DUP #BoardSize + SWAP
			DO
				\ J - pos2
				\ I - pos3
				J I <> I not-empty-at? AND
				IF
					I enemy-at? pos2Enemy @ NOT AND
					I friend-at? OR
					IF
						I enemy-at? pos3Enemy !
						\ Compute pos4			
						DUP 11 MOD J 11 MOD - I + \ pos4
						DUP not-empty-at?
						IF
							DUP enemy-at?
							pos2Enemy @ NOT AND
							pos3Enemy @ NOT AND
							OVER friend-at? OR
							IF
								DUP enemy-at?
								IF
									DUP ringbuffer-add-item		\ pos4
								ELSE
									pos2Enemy @
									IF
										J ringbuffer-add-item	\ pos2
									ELSE
										pos3Enemy @
										IF
											I ringbuffer-add-item	\ pos3
										ENDIF
									ENDIF
								ENDIF
							ENDIF
						ENDIF
						DROP \ Discard pos4
					ENDIF
				ENDIF
			11 +LOOP
		ENDIF
	LOOP
	DROP \ Discard to
;

\ Determine of the position is squared by the opponent.
: SquaredByEnemy? ( to -- ? )
	DUP 11 / 11 * DUP 11 + SWAP
	DO
		\ I - pos2
		DUP I <> I enemy-at? AND
		IF
			I 11 MOD DUP #BoardSize + SWAP
			DO
				\ J - pos2
				\ I - pos3
				J I <> I enemy-at? AND
				IF
					\ Compute pos4			
					DUP 11 MOD J 11 MOD - I +
					enemy-at?
					IF
						R> R> 2DROP \ Exit the I loop.
						R> R> 2DROP \ Exit the J loop.
						DROP \ to
						TRUE EXIT
					ENDIF
				ENDIF
			11 +LOOP
		ENDIF
	LOOP
	DROP \ to
	FALSE
;

\ The "Marks" board keeps track of exchanges in order to avoid duplicates.
CREATE Marks #BoardSize ALLOT

: UnmarkAll
	Marks #BoardSize ERASE
;

: Mark ( index -- )
	Marks + TRUE SWAP C!
;

: Marked? ( index -- ? )
	Marks + C@
;
 
: Unmarked? ( index -- ? )
	Marked? NOT
;


CREATE TempBoard
#BoardSize CELLS ALLOT

\ Save the current board into a temporary location.
: SaveBoard
	$board @ TempBoard #BoardSize MOVE
;

\ Restore the current board from the temporary location.
: RestoreBoard
	TempBoard $board @ #BoardSize MOVE
;

\ Generate the squaring moves and return and indication if any were found.
: SquaringMoves ( from to -- ? )
	SaveBoard
	UnmarkAll	

	ringbuffer-reset

	SWAP board[] #Empty SWAP !	\ Move the piece on the temporary board.
	current-piece OVER board[] !

	CheckSquares			\ Check squaring for the "to" position.

	FALSE				\ No squaring moves found yet.

	BEGIN
		ringbuffer-empty? NOT		\ Pending exchanges?
	WHILE
		ringbuffer-remove-item		\ Get next exchange.

		DUP Marked?		\ Already exchanged?
		IF
			DROP
		ELSE
			current-player OVER change-owner-at
			current-piece  OVER board[] !	\ Change the piece on the temporary board.
			DUP CheckSquares	\ Check squaring due to the exchange.
			Mark			\ Remember this exchange.
			DROP TRUE		\ Indicate squaring move found.
		ENDIF
	REPEAT

	RestoreBoard
;

\ Determine if we are moving into an opponent's square.
: MovingIntoSquare? ( from to -- ? )
	SaveBoard

	SWAP board[] #Empty SWAP !	\ Move the piece.
	current-piece OVER board[] !
	
	SquaredByEnemy?

	RestoreBoard
;

\ Resets the current player's marker back to empty.
: ResetMarker ( -- )
	current-player Red =
	IF
		RedMarker
	ELSE
		YellowMarker
	ENDIF

	DUP not-empty-at?
	IF
		capture-at
	ELSE
		DROP
	ENDIF
;

\ Generate a move in the forward direction.
: ForwardMove ( 'dir -- )
	EXECUTE
	IF
		empty?
		IF
			from here MovingIntoSquare? NOT
			IF
				from here move
				from here SquaringMoves DROP
				ResetMarker
				add-move
			ENDIF
		ENDIF
	ENDIF
;

\ Determine if a "MarkerI" exists at the given position.
: MarkerI-at? ( pos -- ? )
	piece-type-at DUP YellowMarkerI = SWAP RedMarkerI = OR
;

\ Generate a move in the sideways direction.
: SidewaysMove ( 'dir -- )
	EXECUTE
	IF
		empty?
		IF
			current-player Red =
			from here MovingIntoSquare? NOT
			IF
				from here move
				from here SquaringMoves
				IF
					ResetMarker
					add-move
				ELSE
					Referee		
					current-player Red =
					IF	
						RedMarkerI RedMarker
						RedMarkerII >R		
					ELSE
						YellowMarkerI YellowMarker
						YellowMarkerII >R
					ENDIF

					DUP empty-at?
					IF
						create-player-piece-type-at add-move
						R> DROP
					ELSE
						DUP MarkerI-at?
						IF
							SWAP DROP R> SWAP change-type-at add-move DROP
						ELSE \ MarkerII must be present
							R> 2DROP 2DROP
						ENDIF
					ENDIF
				ENDIF
			ENDIF
		ENDIF
	ENDIF
;

: ManMoveNorth		['] North	ForwardMove ;
: ManMoveNortheast	['] Northeast	ForwardMove ;
: ManMoveNorthwest	['] Northwest	ForwardMove ;
: ManMoveEast		['] East	SidewaysMove ;
: ManMoveWest		['] West	SidewaysMove ;


{moves ManMoves
	{move} ManMoveNorth
	{move} ManMoveNortheast
	{move} ManMoveNorthwest
	{move} ManMoveEast
	{move} ManMoveWest
moves}


{pieces
	{piece}		Man		{moves}  ManMoves
	{piece}		RedMarkerI
	{piece}		RedMarkerII
	{piece}		YellowMarkerI
	{piece}		YellowMarkerII
pieces}

(
**********
* Events *
**********
)

\ Return the goal positions for the specfied player.
: GoalPositions ( player -- pos1 pos2 pos3 )
	Red =
	IF
		e11 f11 g11
	ELSE
		e1 f1 g1
	ENDIF
;

(
  This table provides the value of a piece on a given position.  A value
  of 100 indicates the piece value.  The one's digit indicates progress
  towards the goal squares.

  The table is from Red's perspective.  It can be flipped for the opponent
  by subtracting the position from #BoardSize-1.
)
CREATE BoardWeight	\ From Red's perspective
104 , 105 , 108 , 110 , 500 , 500 , 500 , 110 , 108 , 105 , 104 ,
104 , 105 , 108 , 110 , 112 , 114 , 112 , 110 , 108 , 105 , 104 ,
104 , 105 , 108 , 108 , 108 , 108 , 108 , 108 , 108 , 105 , 104 ,
104 , 105 , 105 , 105 , 105 , 105 , 105 , 105 , 105 , 105 , 104 ,
104 , 104 , 104 , 104 , 104 , 104 , 104 , 104 , 104 , 104 , 104 ,
103 , 103 , 103 , 103 , 103 , 103 , 103 , 103 , 103 , 103 , 103 ,
102 , 102 , 102 , 102 , 102 , 102 , 102 , 102 , 102 , 102 , 102 ,
101 , 101 , 101 , 101 , 101 , 101 , 101 , 101 , 101 , 101 , 101 ,
100 , 100 , 100 , 100 , 100 , 100 , 100 , 100 , 100 , 100 , 100 ,
  0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,
  0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,   0 ,

\ Compute the aggregate board weight for both players from Red's perspective.
: AggregateWeight ( -- n )
	0		\ Total weight.
	#BoardSize 0
	DO
		I player-at DUP #Empty <>
		IF
			Red =
			IF
				I CELLS BoardWeight + @ +
			ELSE

				[ #BoardSize 1- ] LITERAL I - CELLS BoardWeight + @ -
			ENDIF
		ELSE
			DROP
		ENDIF
	LOOP
;

\ Evaluation function.
: OnEvaluate
	AggregateWeight
	current-player Yellow = IF NEGATE ENDIF
;

\ Determine if a given player has at least one piece on the board.
: Has3orMorePieces? ( player -- ? )
	0	\ No friendly pieces found yet.
	#BoardSize 0
	DO
		I player-at 2 PICK =
		IF \ Found a friendly piece.
			1+ DUP 2 >
			IF
				LEAVE \ Terminate the loop.
			ENDIF
		ENDIF
	LOOP
	SWAP DROP 2 >
;

\ Determine if a player has reached all three goal positions.
: GoalReached? ( player -- ? )
	GoalPositions
	not-empty-at? SWAP
	not-empty-at? AND SWAP
	not-empty-at? AND
;

\ End of game determination.
: OnIsGameOver ( -- gameResult )
	next-player GoalReached?
	IF
		#LossScore
	ELSE
		\ If it's our turn to move and the other player has fewer than two
		\   pieces left, then we won.
		next-player Has3orMorePieces?
		IF
			\ If it's our turn to move and we have few than two
			\  pieces, then we lost.
			current-player Has3orMorePieces?
			IF
				#UnknownScore
			ELSE
				#LossScore
			ENDIF
		ELSE
			#WinScore
		ENDIF
	ENDIF
;

\ This is required for AutoPlay.
{board-setup
	{setup} Yellow Man b9
	{setup} Yellow Man c9
	{setup} Yellow Man d9
	{setup} Yellow Man e9
	{setup} Yellow Man f9
	{setup} Yellow Man g9
	{setup} Yellow Man h9
	{setup} Yellow Man i9
	{setup} Yellow Man j9

	{setup} Red Man b3
	{setup} Red Man c3
	{setup} Red Man d3
	{setup} Red Man e3
	{setup} Red Man f3
	{setup} Red Man g3
	{setup} Red Man h3
	{setup} Red Man i3
	{setup} Red Man j3
board-setup}
